@@ -122,13 +122,21 @@ class TestFormParser:
         req.max_form_parts = 1
         pytest.raises(RequestEntityTooLarge, lambda: req.form["foo"])
 
-    def test_x_www_urlencoded_max_form_parts(self):
+    def test_urlencoded_no_max(self) -> None:
         r = Request.from_values(method="POST", data={"a": 1, "b": 2})
         r.max_form_parts = 1
 
         assert r.form["a"] == "1"
         assert r.form["b"] == "2"
 
+    def test_urlencoded_silent_decode(self) -> None:
+        r = Request.from_values(
+            data=b"\x80",
+            content_type="application/x-www-form-urlencoded",
+            method="POST",
+        )
+        assert not r.form
+
     def test_missing_multipart_boundary(self):
         data = (
             b"--foo\r\nContent-Disposition: form-field; name=foo\r\n\r\n"
@@ -144,6 +152,32 @@ class TestFormParser:
         )
         assert req.form == {}
 
+    def test_chunk_split_on_line_break_before_epilogue(self):
+        data = b"".join(
+            (
+                # exactly 64 bytes of header
+                b"--thirteenbytes\r\n",
+                b"Content-Disposition: form-data; name=tx3065\r\n\r\n",
+                # payload that fills 65535 bytes together with the header
+                b"\n".join([b"\r" * 31] * 2045 + [b"y" * 31]),
+                # This newline is split by the first chunk
+                b"\r\n",
+                # extra payload that also has the final newline split exactly
+                # at the chunk size.
+                b"\n".join([b"\r" * 31] * 2047 + [b"x" * 30]),
+                b"\r\n--thirteenbytes--",
+            )
+        )
+        req = Request.from_values(
+            input_stream=io.BytesIO(data),
+            content_length=len(data),
+            content_type="multipart/form-data; boundary=thirteenbytes",
+            method="POST",
+        )
+        assert len(req.form["tx3065"]) == (131072 - 64 - 1)
+        assert req.form["tx3065"][-1] == "x"
+        assert req.form["tx3065"][65470:65473] == "y\r\n"
+
     def test_parse_form_data_put_without_content(self):
         # A PUT without a Content-Type header returns empty data
 
@@ -166,6 +200,14 @@ class TestFormParser:
         assert len(form) == 0
         assert len(files) == 0
 
+    def test_parse_form_post_data_trailing_CR(self):
+        for k in [1, 2]:
+            sample = b"\0" * 65535 + b"\x0d" * k
+            with Request.from_values(
+                data={"foo": (io.BytesIO(sample), "test.txt")}, method="POST"
+            ) as req:
+                assert req.files["foo"].read() == sample
+
     @pytest.mark.parametrize(
         ("no_spooled", "size"), ((False, 100), (False, 3000), (True, 100), (True, 3000))
     )
@@ -197,7 +239,8 @@ class TestFormParser:
         assert stream is not None
 
 
-@pytest.mark.filterwarnings("ignore::pytest.PytestUnraisableExceptionWarning")
+# TODO Fix the ResourceErrors, somewhere in here a SpooledTemporaryFile is not
+#   getting closed. This causes PytestUnraisableExceptionWarning.
 class TestMultiPart:
     def test_basic(self):
         resources = join(dirname(__file__), "multipart")
@@ -448,3 +491,15 @@ class TestMultiPartParser:
         ) as request:
             assert request.files["rfc2231"].filename == "a b c d e f.txt"
             assert request.files["rfc2231"].read() == b"file contents"
+
+
+def test_multipart_max_form_memory_size() -> None:
+    """max_form_memory_size is tracked across multiple data events."""
+    data = b"--bound\r\nContent-Disposition: form-field; name=a\r\n\r\n"
+    data += b"a" * 15 + b"\r\n--bound--"
+    # The buffer size is less than the max size, so multiple data events will be
+    # returned. The field size is greater than the max.
+    parser = formparser.MultiPartParser(max_form_memory_size=10, buffer_size=5)
+
+    with pytest.raises(RequestEntityTooLarge):
+        parser.parse(io.BytesIO(data), b"bound", None)
